import os
import subprocess
from subprocess import PIPE
import boto3
import csv
import glob
import time
import datetime
from modules import ledhandler

REGION_NAME = 'ap-northeast-1'
CONFIG_PATH = '/etc/monitoring_camera/'
DEFAULT_SAVE_PATH = '/opt/monitoring_camera/pictures'

class AWS():
    def __init__(self):
        os.environ['AWS_DEFAULT_REGION'] = REGION_NAME
        self._set_credentials_to_env()

        self.s3 = self.S3()
        self.dynamodb = self.DynamoDB()

        self._deploy_to_ec2()

    def _deploy_to_ec2(self):
        pems = glob.glob(os.path.join(CONFIG_PATH, '*.pem'))
        if len(pems) != 1:
            return
        pem = pems[0]
        os.chmod(pem, 0o400)

        ec2_files = os.path.join(CONFIG_PATH, 'ec2_files')
        s3_bucket_name = os.getenv('S3_BUCKET_NAME')
        ec2_user_name = os.getenv('EC2_USER_NAME')
        ec2_public_dns= os.getenv('EC2_PUBLIC_DNS')
        self._replace_txt(
            file=os.path.join(ec2_files, 'flask', 'modules', 'awsinteract.py'),
            before="TABLE_NAME = ''",
            after="TABLE_NAME = '{}'".format(s3_bucket_name)
        )

        # 対象のEC2インスタンス上に既にFlaskアプリがデプロイされていたらなにもしない
        cmd = 'ssh -i "{}" -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null {}@{} [ -e /var/www/flask ] && echo OK'.format(
            pem,
            ec2_user_name,
            ec2_public_dns
        )
        res = subprocess.run(cmd, stdout=PIPE, stderr=PIPE, shell=True)
        if 'OK' in res.stdout.decode():
            return

        # デプロイ用のファイルをEC2に送信
        cmd = 'scp -r -i "{}" -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null {} {}@{}:~/'.format(
            pem,
            ec2_files,
            ec2_user_name,
            ec2_public_dns
        )
        res = subprocess.run(cmd, stdout=PIPE, stderr=PIPE, shell=True)

        # EC2上で送信したデプロイスクリプトを動かす
        cmd = 'ssh -i "{}" -o StrictHostKeyChecking=no -o UserKnownHostsFile=/dev/null {}@{} sudo sh ec2_files/deploy.sh {}'.format(
            pem,
            ec2_user_name,
            ec2_public_dns,
            ec2_user_name
        )
        res = subprocess.run(cmd, stdout=PIPE, stderr=PIPE, shell=True)

    def _replace_txt(self, file, before, after):
        with open(file, 'r') as f:
            data_lines = f.read()

        data_lines = data_lines.replace(before, after)

        with open(file, 'w') as f:
            f.write(data_lines)

    def _set_credentials_to_env(self):
        with open(os.path.join(CONFIG_PATH, 'credentials.csv'), 'r') as scenario:
            header = next(csv.reader(scenario))
            scenario = csv.reader(scenario)
            for step in scenario:
                for col_type, value in zip(header, step):
                    if col_type == 'Access key ID':
                        accesskeyid = value
                        os.environ['AWS_ACCESS_KEY_ID'] = accesskeyid
                    elif col_type == 'Secret access key':
                        secretaccesskey = value
                        os.environ['AWS_SECRET_ACCESS_KEY'] = secretaccesskey
                    elif col_type == 'Console login link':
                        accountId = value.split('.')[0].split('/')[-1]

    def upload(self):
        period = int(os.getenv('UPLOAD_PERIOD'))
        auto_remove = os.getenv('AUTO_REMOVE')
        while True:
            time.sleep(period)
            file_path = self.s3.upload(auto_remove)
            self.dynamodb.upload(file_path)

    class S3:
        def __init__(self):
            self.s3 = boto3.client('s3')

            # S3バケットの名前を設定
            # ユーザー未指定の場合、「monitoring-camera-[MACアドレス]」となる
            self.bucket_name = os.getenv('S3_BUCKET_NAME')
            if self.bucket_name == '':
                self.bucket_name = self._make_s3_bucket_name()
                os.environ['S3_BUCKET_NAME'] = self.bucket_name

            # 画像の保存先ディレクトリを設定
            # ユーザー未指定、またはユーザー指定のディレクトリを/mntにマウントできなかった場合、デフォルトのディレクトリを使用する
            self.path = ''
            p = os.getenv('SAVE_PATH')
            if p != '':
                cmd = 'mount ' + p + ' /mnt'
                res = subprocess.run(cmd.split(), stdout=PIPE, stderr=PIPE)
                if res.returncode == 0:
                    self.path = '/mnt'
            if self.path != '/mnt':
                self.path = DEFAULT_SAVE_PATH
                os.environ['AUTO_REMOVE'] = 'true'
            os.environ['SAVE_PATH'] = self.path

            # 使用するS3バケットの存在確認(なければ作る)
            if not self._exists_bucket():
                self._create_bucket()

        def _get_armadillo_mac(self):
            return subprocess.getoutput(
                """ip link show dev eth0 | grep link | awk '{print toupper($2)}' | tr -d ':'"""
            )

        def _make_s3_bucket_name(self):
            mac = self._get_armadillo_mac()
            return ('monitoring-camera-' + mac).lower()
            
        def _exists_bucket(self):
            try:
                buckets = self.s3.list_buckets()['Buckets']
            except Exception as e:
                ledhandler.cloud_error()

            for bucket in buckets:
                if self.bucket_name == bucket['Name']:
                    return True
            return False

        def _create_bucket(self):
            try:
                self.s3.create_bucket(
                    Bucket=self.bucket_name,
                    CreateBucketConfiguration={
                        'LocationConstraint': REGION_NAME
                    }
                )
            except (self.s3.exceptions.BucketAlreadyExists,
                    self.s3.exceptions.BucketAlreadyOwnedByYou):
                ledhandler.cloud_error()
            else:
                waiter = self.s3.get_waiter('bucket_exists')
                waiter.wait(Bucket=self.bucket_name)
                self._put_lifecycle()

        def _put_lifecycle(self):
            try:
                self.s3.put_bucket_lifecycle_configuration(
                    Bucket=self.bucket_name,
                    LifecycleConfiguration={
                        'Rules':[
                            {
                                "Expiration": {
	                            "Days": 1
                                },
                                "ID": "rule",
                                "Prefix": "",
                                "Status": "Enabled",
                                "NoncurrentVersionExpiration": {
	                            "NoncurrentDays": 1
                                }
                            }
                        ]
                    }
                )
            except:
                ledhandler.cloud_error()

        def upload(self, auto_remove='true'):
            file_path = []
            # path内の最新版を除くjpgファイルを全てS3に保存
            files = glob.glob(os.path.join(self.path, "**", "*.jpg"), recursive=True)[:-1]
            for file in files:
                file_name = file.split('/')[-1]
                folder_name = file.split('/')[-2]
                p = folder_name + '/' + file_name
                try:
                    self.s3.upload_file(
                        file,
                        self.bucket_name,
                        p,
                        ExtraArgs={
                            "ContentType": "image/jpeg",
                            "ACL": "public-read"
                        }
                    )
                except boto3.exceptions.S3UploadFailedError:
                    ledhandler.cloud_error()
                else:
                    # 保存したファイルはArmadilloから削除する
                    file_path.append(p)
                    if auto_remove == 'true':
                        os.remove(file)

            return file_path

    class DynamoDB:
        def __init__(self):
            self.dynamodb = boto3.client('dynamodb')
            # テーブル名にはS3のバケット名と同一のものを使用する
            self.table_name = os.getenv('S3_BUCKET_NAME')
            if not self._exists_table():
                try:
                    self._create_table()
                except Exception as e:
                    ledhandler.cloud_error()

        def _exists_table(self):
            try:
                tables = self.dynamodb.list_tables()['TableNames']
            except Exception as e:
                ledhandler.cloud_error()

            if self.table_name in tables:
                return True
            else:
                return False

        def _create_table(self):
            table = self.dynamodb.create_table(
                TableName=self.table_name,
                KeySchema=[
                    {
                        'AttributeName': 'identifier',
                        'KeyType': 'HASH' # partition key
                    },
                    {
                        'AttributeName': 'ttl',
                        'KeyType': 'RANGE' # sort key
                    }
                ],
                AttributeDefinitions=[
                    {
                        'AttributeName': 'identifier',
                        'AttributeType': 'S'
                    },
                    {
                        'AttributeName': 'ttl',
                        'AttributeType': 'N'
                    }
                ],
                ProvisionedThroughput={
                    'ReadCapacityUnits': 5,
                    'WriteCapacityUnits': 5
                }
            )

            # テーブルが作成されるまで待つ
            waiter = self.dynamodb.get_waiter('table_exists')
            waiter.wait(TableName=self.table_name)

            # TTLの設定
            result = self.dynamodb.update_time_to_live(
                TableName=self.table_name,
                TimeToLiveSpecification={
                    'Enabled': True,
                    'AttributeName': 'ttl'
                }
            )

        def _create_url(self, bucket_name, filepath):
            # https://[bucket_name].s3-[aws_region].amazonaws.com/[file_name]
            return ('https://' + bucket_name + '.s3-' + os.getenv('AWS_DEFAULT_REGION') + '.amazonaws.com/' + filepath)

        def upload(self, file_path):
            for p in file_path:
                cameraname = p.split('/')[0]
                filename = p.split('/')[1]

                try:
                    day = datetime.datetime.strptime(filename.split('.')[0], '%Y%m%d%H%M%S')
                except Exception as e:
                    ledhandler.cloud_error()

                # ttl(Time To Live を1日後に設定)
                ttl = str(int((day + datetime.timedelta(days=1)).timestamp()))

                url = self._create_url(self.table_name, p)

                try:
                    self.dynamodb.put_item(
                        TableName=self.table_name,
                        Item={
                            'identifier': {'S': cameraname},
                            'ttl': {'N': ttl},
                            'url': {'S': url}
                            }
                        )
                except Exception as e:
                    ledhandler.cloud_error()
