import os
import time
import datetime
import threading
from pyueye import ueye
from PIL import Image
from modules import ledhandler

def get_list_of_cameras():
    n = ueye.is_GetNumberOfDevices()
    if n == 0:
        ledhandler.camera_error()
    return n

class Camera:
    def __init__(self, period, save_path):
        self.period = period
        try:
            self._init_camera()
        except CameraError:
            ledhandler.camera_error()
        self.name = 'Camera_' + self._get_serial_number()
        self.path = os.path.join(save_path, self.name)

    def _init_camera(self):
        self.hCam = ueye.HIDS(0)
        self.pcImageMemory = ueye.c_mem_p()
        self.MemID = ueye.int()
        self.rectAOI = ueye.IS_RECT()
        self.pitch = ueye.INT()
        self.nBitsPerPixel = ueye.INT(24)
        self.m_nColorMode = ueye.INT()
        self.bytes_per_pixel = int(self.nBitsPerPixel / 8)

        nRet = ueye.is_InitCamera(self.hCam, None)
        if nRet != ueye.IS_SUCCESS:
            raise CameraError('is_InitCamera ERROR')

        nRet = ueye.is_ResetToDefault(self.hCam)
        if nRet != ueye.IS_SUCCESS:
            raise CameraError('is_ResetToDefault ERROR')

        nRet = ueye.is_SetDisplayMode(self.hCam, ueye.IS_SET_DM_DIB)

        self.bytes_per_pixel = int(self.nBitsPerPixel / 8)
        self.m_nColorMode = ueye.IS_CM_RGB8_PACKED

        nRet = ueye.is_AOI(self.hCam,
                           ueye.IS_AOI_IMAGE_GET_AOI,
                           self.rectAOI,
                           ueye.sizeof(self.rectAOI))
        if nRet != ueye.IS_SUCCESS:
            raise CameraError('is_AOI ERROR')

        self.width = self.rectAOI.s32Width
        self.height = self.rectAOI.s32Height

        nRet = ueye.is_AllocImageMem(self.hCam,
                                     self.width,
                                     self.height,
                                     self.nBitsPerPixel,
                                     self.pcImageMemory,
                                     self.MemID)
        if nRet != ueye.IS_SUCCESS:
            raise CameraError('is_AllocImageMem ERROR')
        else:
            nRet = ueye.is_SetImageMem(self.hCam,
                                       self.pcImageMemory,
                                       self.MemID)
            if nRet != ueye.IS_SUCCESS:
                raise CameraError('is_SetImageMem ERROR')
            else:
                nRet = ueye.is_SetColorMode(self.hCam, self.m_nColorMode)

    def _get_serial_number(self):
        pInfo = ueye.CAMINFO()
        nRet = ueye.is_GetCameraInfo(self.hCam, pInfo)
        if nRet != ueye.IS_SUCCESS:
            raise CameraError('is_GetCameraInfo ERROR')

        return str(pInfo.SerNo.decode())

    def _take_picture(self):
        nRet = ueye.is_FreezeVideo(self.hCam, ueye.IS_WAIT)
        if nRet != ueye.IS_SUCCESS:
            raise CameraError('is_FreezeVideo ERROR')

        nRet = ueye.is_InquireImageMem(self.hCam,
                                       self.pcImageMemory,
                                       self.MemID,
                                       self.width,
                                       self.height,
                                       self.nBitsPerPixel,
                                       self.pitch)
        if nRet != ueye.IS_SUCCESS:
            raise CameraError('is_InquireImageMem ERROR')

        array = ueye.get_data(self.pcImageMemory,
                              self.width,
                              self.height,
                              self.nBitsPerPixel,
                              self.pitch,
                              copy=False)

        pil_image = Image.fromarray(array.reshape(self.height.value,
                                                  self.width.value,
                                                  self.bytes_per_pixel),
                                    'RGB')

        # save snapshot image
        pil_image.save(os.path.join(self.path,
                                    datetime.datetime.now().strftime("%Y%m%d%H%M%S") + ".jpg"))

    def start_take_picture(self):
        while True:
            time.sleep(self.period)
            try:
                self._take_picture()
            except CameraError:
                ledhandler.camera_error()

    def get_save_path(self):
        return self.path

class CameraError(Exception):
    """ カメラの操作中にエラーが発生した場合に知らせる例外クラス """
    pass
